; HIMEM-Extension to UMBs for DOS 5-7
; Original (c) Peter Siering, published in c't 8/1991
; Extended version by as (Schpers) 15-AUG-1991.
; Extended version by as (Stiller), c't 11/1995
; (Comments translated by Eric Auer 11/2002)
; Assemble/Link with TASM + TLink + Exe2bin
; For Saturn, Mercury, Neptun (C000h-EFFFh)
; Triton (with DMA only E000h-EFFFh !)
; Call as:

; DEVICE=UMBPCI.SYS /I=aaaa-bbbb /I=cccc-dddd....

; Example: DEVICE=UMBPCI.SYS /I=CC00-CFFF /I=E000-EFFF

          .MODEL SMALL
          .CODE
          .386

INITREQUEST STRUC   ; parameter block for INIT
  irLength        db ?    ; total size of this parameter block
  irUnit          db ?    ; not used
  irFunction      db ?    ; function $00, (init)
  irStatus        dw ?    ; return status of initialization
  irReserved      db 8 dup (?) ; "reserved", not used
  irUnits         db ?    ; number of units handled by this driver
  irEndAddress    dd ?    ; highest available memory address
  irParamAddress  dd ?    ; pointer to the command line (device=...)
  irDriveNumber   db ?    ; first free drive number
  irMessageFlag   db ?    ; flags for error messages
INITREQUEST ENDS

; Gertetreiberkopf
DNext     dd -1           ; address of the next driver: FFFF:FFFF
DAttr     dw 0E000h       ; attributes
DStrat    dw OFFSET Strat ; offset of the "strategy"-handler
DIntr     dw OFFSET Intr  ; offset of the "interrupt"-handler
DName     db 'UMBADDXX'   ; name (or units(?))

ParamAddr    dd ?         ; buffer for init parameter block

; strategy-handler, saves ES:BX (address of the parameter block)
Strat     PROC FAR
          mov Word Ptr [ParamAddr],bx
          mov Word Ptr [ParamAddr+2],es
          retf
Strat     ENDP

; interrupt-handler, defines only the INIT-function
Intr      PROC FAR
          push ax         ; we only need to save those
          push bx         ; 3 registers for now
          push es
          pushf
          les  bx,[ParamAddr] ; ES:BX points to the parameter block
          mov  es:[bx+irStatus],8103h ; pessimistic approach
          mov  al,es:[bx+irFunction] ; funktion number
          or   al,al      ; Init?
          jnz  OtherCmd   ; return status "unknown command"
          call InitFunc   ; INIT - returns status in AX
          les  bx,[ParamAddr]
          mov  es:[bx+irStatus],ax
OtherCmd:
	  popf
          pop  es
          pop  bx
          pop  ax
          retf
Intr      ENDP

UMBEntry STRUC
Start     dw ?      ; start address (paragraphs)
Len       dw ?      ; size in paragraphs
UMBEntry ENDS

UMBList  dw 22 dup(0) ; max. 10 entries!

; is inserted into the chain by IntFunc and called by DOS
; to allocate UMBs.
; Call:    DX = needed size (bytes), flags on stack
; Returns: DX = size of the found block (bytes), BX = segment, AX = 1
;          or AX = 0 and BL = $B1 for "no block at all"
;          or AX = 0 and BL = $B0 for "block too small"
UMBRequest PROC NEAR
          push ds
          push si
          push cx
          push bx

          push cs
          pop  ds     ; DS as CS
          mov  si,OFFSET UMBList
          xor  cx,cx  ; for size comparison
          cld         ; count upwards
UMBSearch:
          lodsw       ; one entry of the UMB list
          or   ax,ax  ; end of the list?
          jz   UMBEnd ; if yes, not found
          mov  bx,ax  ; remember start segment
          lodsw
          cmp  bx,-1  ; already used?
          jz   UMBSearch ; if yes, next entry
          cmp  ax,cx  ; biggest block yet?
          jb   ChkSize
          mov  cx,ax  ; if yes, remember size
ChkSize:  cmp  ax,dx  ; size big enough?
          jb   UMBSearch ; if no, next block
UMBFound: mov Word Ptr[si-4],-1 ; mark block as used
          mov  dx,ax  ; size in DX, segment in BX
          pop  ax     ; old BX from stack
          mov  ax,1   ; signal success
          jmp  Short UMBDone
UMBEnd:   xor  ax,ax
          pop  bx     ; BX-original (BH)
          mov  bl,0B0h ; assumption: block only too small
          mov  dx,cx   ; returned block size
          or   dx,dx  ; found anything?
          jnz  UMBDone ; if yes, only too small
          inc  bl     ; found nothing at all: BL = $B1
UMBDone:  pop  cx
          pop  si
          pop  ds
          popf
          retf
UMBRequest ENDP

; chain of HIMEM.SYS
XMSOrgJmp dd ?       ; original value
NewChain: jmp short NewCtrl
          nop        ; the next handler would use this
          nop        ; space to patch in its jump
          nop
NewCtrl:  pushf
          cmp  ah,10h  ; UMB request?
          jz  UMBRequest ; flags on stack!
          popf
          jmp  [XMSOrgJmp] ; else chain on

;
END_RESIDENT LABEL   BYTE   ; end of resident part

; *******************************************
; *** start of non resident (loader) part ***
; *******************************************

CRLF      db 13,10,'$'

PrintString:
          push dx
          mov  dx,OFFSET CRLF
          call DoPrint
          pop  dx
          call DoPrint
          mov  dx,OFFSET CRLF
DoPrint:  mov  ah,09h
          int  21h
          ret

; print a 4-digit hex value (AX)
HDigits   db '0123456789ABCDEF'
PrHex:    mov  cx,0Ch   ; shift 12 bits in the first round
PrDig:    push ax
          shr  ax,cl
          and  ax,000Fh
          mov  bx,ax
          mov  dl,[HDigits+bx]
          mov  ah,2     ; "display output"
          int  21h
          pop  ax
          sub  cl,4
          jnb  PrDig
          ret

; Print an error message using DOS function 9 and exit
; Call with DS:DX = pointer to error message
ErrOut:   call PrintString               ; print string
          mov  ax,0
          int  16h
          les  bx,[ParamAddr]
          mov  Word Ptr es:[bx+irEndAddress],0 ; 0 bytes resident size
          mov  ax,8100h  ; error (unspecific)
          jmp  InitEnd

; Initialization - called with ES:DI = parameter block and
; ES,BX,AX and flags on stack, i.e. does not have to save those registers
InitFunc  PROC NEAR
          push cx
          push dx
          push si
          push di
          push ds

          push cs
          pop  ds          ; DS = CS

          mov  Word Ptr es:[bx+irEndAddress],OFFSET CRLF
          mov  Word Ptr es:[bx+irEndAddress+2],cs ; end of resident part
          mov  dx,OFFSET Copyright$  ; "hello world"
          call PrintString

; HIMEM installed?
chkHimem:
          mov  ax,4300h    ; install check of HIMEM
          int  2Fh
          cmp  al,80h      ; installed?
          jz   GetXMSCall  ; if yes...
          mov  dx,OFFSET NoXMS$
          jmp  ErrOut
; find chain
GetXMSCall:
          mov  ax,4310h    ; get HIMEM handler address
          int  2Fh
          mov  Word Ptr [XMSOrgJmp],bx
          mov  Word Ptr [XMSOrgJmp+2],es
; UMB server already existing?
          mov  ah,16
          mov  dx,0FFFh    ; allocate a 64k block
          call [XMSOrgJmp]
          cmp  bl,80h      ; not installed?
          jz   GetCmdLine  ; then okay
          mov  dx,OFFSET UMBPresent$
          jmp  ErrOut      ; else error, UMB already existing
; process command line
GetCmdLine:
          les  di,[ParamAddr]  ; parameter block
          les  di,es:[di+irParamAddress] ; ES:DI points to the command line
          call SetList     ; parse and set list
          jnc  TestAreas   ; if no errors, go on
          mov  dx,OFFSET BadCmd$
          jmp  ErrOut      ; else print command line and error message

TestAreas:CALL INITChipset
          JC   chipseterr

          mov  si,OFFSET UMBList
          cld
NextUMB:  lodsw            ; starting address
          or  ax,ax        ; end of the list reached?
          jz  AreasTested
          mov es,ax        ; starting address in ES
          lodsw            ; size (paragraphs)
          mov  cl,10       ; convert to 16K units
          shr  ax,cl       ; => number of 16k blocks
TstBlocks: xor  di,di
          Call setchipset   ; enable segment ES, size 16
          jc   Chipseterr
          mov  bx,es:[di]   ; read current content
          mov  dx,bx
          xor  dx,0FFFFh    ; flip
          mov  es:[di],dx   ; and write
          mov  cx,80h       ; wait a moment in case the operation
          loop $            ; ended up being a no-op
          cmp  es:[di],dx   ; do we have a constant value?
          mov  es:[di],bx   ; write back old value
          jnz  NoRAM        ; -> no RAM found at this place!
          dec  ax
          jz   NextUMB      ; -> next UMB area
          mov  di,es
          add  di,0400h     ; advance ES by 16K
          mov  es,di
          jmp  TstBlocks    ; check next block

NoRAM:    push es           ; segment address of our block
          mov  dx,OFFSET RAMFail1$
          call DoPrint      ; print without CRLF
          pop  ax
          call PrHex        ; print segment address
          mov  dx,OFFSET RAMFail2$
          jmp  ErrOut

chipseterr:mov dx,OFFSET SetErr$
           jmp ErrOut

; Here we continue after successful RAM check and hooking of the
; HIMEM.SYS chain
AreasTested:
           les di,[XMSOrgJmp]
ScanHooks: mov  al,Byte Ptr es:[di]
          cmp  al,0EBh      ; JMP SHORT?
          jz   HookFound    ; OK - chain currently ends here
          les  di,DWord Ptr es:[di+1] ; else try next element
          cmp  al,0EAh      ; is this a far jump after all?
          jz   ScanHooks
          mov  dx,OFFSET HookErr$  ; else we cannot follow the chain, error
          jmp  ErrOut
HookFound: cli               ; please do not disturb!
          dec  Byte Ptr es:[di] ; JMP SHORT -> JMP FAR
          mov  Word Ptr es:[di+1], OFFSET NewChain ; activate own
          mov  Word Ptr es:[di+3], cs   ; handler
          add  di,5         ; real original address (after the NOPS)
          mov  Word Ptr [XMSOrgJmp],di
          mov  Word Ptr [XMSOrgJmp+2],es
          sti

showinst:
          mov  dx,OFFSET Installed$
          call PrintString
          mov  ax,0100h    ; Initializing successful
InitEnd:  pop  ds
          pop  di
          pop  si
          pop  dx
          pop  cx
          ret
InitFunc  ENDP

Copyright$  db 'UMBADD c''t 11/95 - Siering/Schpers/Stiller$'
NoXMS$      db 'No XMS driver installed!$'
UMBPresent$ db 'UMB server already installed!$'
BadCmd$    db 'Problems with command line syntax in CONFIG.SYS!',13,10
           db 'Example: DEVICE=UMBADD.SYS /Iaaaa-bbbb /Icccc-dddd ...$'
RAMFail1$   db 13,10,'No RAM in segment (hex) $'
RAMFail2$   db '- installation aborted -$'
HookErr$    db 'Hook-chain of HIMEM.SYS damaged or unknown!$'
Installed$  db 'Program installed.$'
SetErr$     db 'Problems with PCI-chipset-programming!$'

; Call with ES:DI pointing to the command line
; returns with DI+4, value or 0 in DX and clear/set zero-flag
HexToVal: push cx
          mov  cx,0404h     ; counter and shift counter
          xor  dx,dx        ; value
DoDigit:  shl  dx,cl        ; one digit to the left
          mov  al,es:[di]   ; one character
          inc  di
          cmp  al,'9'
          jbe  IsDigit
          and  al,0DFh      ; uppercase letters
          cmp  al,'A'
          jb   NoChar       ; error
          cmp  al,'F'
          ja   NoChar       ; dito
          sub  al,55        ; 'A'..'F' -> 10..16
          jmp  short SetDig ; insert
IsDigit:  cmp  al,'0'
          jb   NoChar       ; error
          sub  al,'0'       ; '0'..'9' -> 0..9
SetDig:   or   dl,al        ; insert
          dec  ch
          jnz  DoDigit
          jmp  Short DigDone
NoChar:   xor  dx,dx        ; error code: dx = 0
DigDone:  pop  cx
          or   dx,dx
          ret

; Parse the command line. Call with ES:DI pointing to the
; command line. Returns with set UMBList and set/clear carry-flag
SetList   PROC NEAR
          xor  bx,bx        ; index in UMBList
DoElem:   mov  al,es:[di]
          inc  di
          cmp  al,13        ; CR?
          jz   ListSet      ; if yes, end
          cmp  al,10        ; LF?
          jz   ListSet      ; -> dito
          cmp  al,'/'       ; slash of "/I=..."?
          jnz  DoElem
          mov  ax,es:[di]
          and  al,0DFh      ; translate 'i' -> 'I'
          cmp  ax,'=I'      ; is "I="? (sic!)
          jnz  listerr       ; if no, TestC
          add  di,2         ; if yes, skip
          call HexToVal     ; parse hex number -> DX
          jz   ListErr      ; -> failed
          mov  al,es:[di]
          cmp  al,'-'       ; '-' ?
          inc  di
          mov  cx,dx        ; store starting address
          call HexToVal     ; get end address
          jz   ListErr      ; -> failed
          mov  Word Ptr[UMBList+bx],cx  ; store starting address
          mov  ax,dx
          sub  ax,cx        ; size in paragraphs
          inc  ax           ; B000-B7FF are $7FF + 1 paragraphs!
          mov  Word Ptr[UMBList+bx+2],ax
          add  bx,4
          cmp  bx,40        ; more than 10 entries?
          jb   DoElem       ; if not, loop on

ListErr:  stc
          ret
ListSet:  sub  bx,1         ; set carry if BX is 0
          ret
SetList   ENDP

;**********************************************************
;************* from here on PCI-chipset-programming *******
;**********************************************************

PCI      Macro n
         MOV   AH,0B1h
         MOV   AL,n
         INT   1Ah
         endm

Liste     DW SIOlist;
BUSDEV    DW 0;

RWC_Lo  EQU  0F0Fh
RWC_Hi  EQU 0F0F0h

SIOlist:  ; segments and register values for Saturn, Mercury, Neptun
 dw 0C000h, 12h, 5Ah,RWC_lo, 154h,0303h,  149h,0101h, 0
 dw 0C400h, 12h, 5Ah,RWC_hi, 154h,0C0Ch,  149h,0202h, 0
 dw 0C800h, 12h, 5Bh,RWC_lo, 154h,3030h,  149h,0404h, 0
 dw 0CC00h, 12h, 5Bh,RWC_hi, 154h,0C0C0h, 149h,0808h, 0
 dw 0D000h, 12h, 5Ch,RWC_lo, 155h,0303h,  149h,1010h, 0
 dw 0D400h, 12h, 5Ch,RWC_hi, 155h,0C0Ch,  149h,2020h, 0
 dw 0D800h, 12h, 5Dh,RWC_lo, 155h,3030h,  149h,4040h, 0
 dw 0DC00h, 12h, 5Dh,RWC_hi, 155h,0C0C0h, 149h,8080h, 0
 dw 0E000h, 16h, 5Eh,RWC_lo, 156h,0303h,  148h,0808h, 14Eh,0040h, 0
 dw 0E400h, 16h, 5Eh,RWC_hi, 156h,0C0Ch,  148h,0808h, 14Eh,0040h, 0
 dw 0E800h, 16h, 5Fh,RWC_lo, 156h,3030h,  148h,0808h, 14Eh,0040h, 0
 dw 0EC00h, 16h, 5Fh,RWC_hi, 156h,0C0C0h, 148h,0808h, 14Eh,0040h, 0
 Dw 00000h

PIIXlist:   ; segments and register values for Triton
 dw 0C000h, 0Ah, 5Ah,RWC_lo, 0   ; no ISA-DMA possible !!!
 dw 0C400h, 0Ah, 5Ah,RWC_hi, 0   ; no ISA-DMA possible !!!
 dw 0C800h, 9Ah, 5Bh,RWC_lo, 0   ; no ISA-DMA possible !!!
 dw 0CC00h, 0Ah, 5Bh,RWC_hi, 0   ; no ISA-DMA possible !!!
 dw 0D000h, 0Ah, 5Ch,RWC_lo, 0   ; no ISA-DMA possible !!!
 dw 0D400h, 0Ah, 5Ch,RWC_hi, 0   ; no ISA-DMA possible !!!
 dw 0D800h, 0Ah, 5Dh,RWC_lo, 0   ; no ISA-DMA possible !!!
 dw 0DC00h, 0Ah, 5Dh,RWC_hi, 0   ; no ISA-DMA possible !!!
 dw 0E000h, 12h, 5Eh,RWC_lo, 169h,0808h, 14Eh,0040h, 0
 dw 0E400h, 12h, 5Eh,RWC_hi, 169h,0808h, 14Eh,0040h, 0
 dw 0E800h, 12h, 5Fh,RWC_lo, 169h,0808h, 14Eh,0040h, 0
 dw 0EC00h, 12h, 5Fh,RWC_hi, 169h,0808h, 14Eh,0040h, 0
 Dw 00000h

setpci    Proc near              ; SI=>list, BX= BUSDEV
          MOV  SI,liste          ; register-list
 @Sloop:  MOV  AX,Word Ptr [SI]
      	  OR   AX,AX
          STC
          JZ   @fertig            ; exit with carry
          MOV  DX,ES
          CMP  DX,AX
          JZ   @found
          ADD  SI,Word PTR [SI+2]
          JMP  @Sloop
 @found:
          MOV DI,44h              ; Master Enable
          MOV CL,17h              ; PCIMaster: BIOS:Ro, 8000-A000:rw
          MOV BX,BUSDEV
          PCI 0Bh                 ; write
 @floop:  ADD  SI,4
	  MOV  AX, word ptr [SI]  ; DEV+func+register -> AX
          OR   AX,AX
          JZ   @fertig
          XOR  BX,BX               ;
	  XCHG BL,AH               ;
          MOV  DI,AX               ; register -> DI
          CMP  BL,1                ; code for: take [BUSDEV]
          JNZ  @set1               ;
          MOV  BX,BUSDEV           ; bus+DEV+func
@set1:    PCI  08h                 ; content -> CL
          JC   @fertig
          MOV  AL, Byte Ptr [SI+2] ; AND mask = Not Lobyte
          NOT  AL
          AND  CL, AL
	  OR   CL, Byte Ptr [SI+3] ; OR mask  = HiByte
          PCI  0Bh
          JNC  @floop
@fertig:  RET
setpci    endp

InitChipset PROC Near
@CheckSIO:MOV DX,8086h             ; INTEL
          MOV CX,0484h             ; SIO-chip
          MOV SI,0
          PCI 02
          JNC @InitExit
          MOV Liste, offset SIOList
@CheckPIIX:MOV DX,122Eh              ; Triton-PIIX
          PCI 02
          JC  @Initexit
          MOV Liste, offset PIIXlist
@Initexit:MOV BUSDEV,BX
          RET
InitChipset Endp

SetChipset proc near
          pusha
	  CALL SetPCI
          JC   @setexit
;Initialize memory if board works with parity
          mov di,0       ; turn off parity check
          mov dx,61h
          in  al,dx      ; remember old setting
          push ax
	  or  al,0Ch
          out dx,al
          mov cx,1000h   ; fill memory with 0
          mov eax,0
          rep stosd
          in  al,dx      ; reset parity-flipflop
          pop ax
          out dx,al      ; back to the old setting
	  popa
          clc
@Setexit: ret
Setchipset endp

  	  END

